/* -*- mode: c++; c-basic-offset: 3; -*- */
#include <algorithm>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <string>
#include <iostream>
#include <fstream>
#include <TGFSComboBox.h>
#include <TGListView.h>
#include <TGTextView.h>
#include <TGTextEdit.h>
#include <TGFSContainer.h>
#include <TGMsgBox.h>
#include <TSystem.h>
#include "filterwiz/TLGFilterWizard.hh"
#include "filterwiz/TLGFilterDlg.hh"
#include "filterwiz/FilterFile.hh"
#include "PlotSet.hh"
#include "ligogui/TLGEntry.hh"
#include "ligogui/TLGTextEditor.hh"
#include "ligogui/TLGComboEditBox.hh"
#include "ligogui/TLGPad.hh"
#include "iirutil.hh"
#include "constant.hh"

   using namespace std;
   using namespace ligogui;

//______________________________________________________________________________
extern "C" {
   // These function is needed by the FilterDesign class
   // These function is needed by DTT to edit a filter string
   bool wizard___dynamic (const std::string name, std::string& filter);
}

namespace filterwiz {


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Constants							        //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   const char* const aboutmsg = 
   "LIGO))) Laser Interferometer Gravitational-wave Observatory\n"
   "Foton - Filter Online Tool\n\n"
   "by Daniel Sigg, 2002, 2003\n"
   "version " VERSION "\n\n"
   "http://www.ligo.caltech.edu\n"
   "http://www.ligo-wa.caltech.edu/gds";

   static const char* const gSaveAsTypes[] = { 
   "Filter files", "*.txt",
   "All files", "*",
   0, 0 };


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Widget ID numbers						        //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   const int kPathId = 100;
   const int kFileId = 101;
   const int kModuleId = 102;
   const int kDirUpId = 103;
   const int kSectMulId = 120;
   const int kSectSelAllId = 121;
   const int kSectSelNoneId = 122;
   const int kSectId = 123;
   const int kSectNameId = 133;
   const int kSwitchInpId = 150;
   const int kSwitchOutId = 151;
   const int kSwitchParamId = 152;
   const int kPlotParamId = 160;
   const int kPlotTypeId = 164;
   const int kSampleId = 170;
   const int kCommandId = 171;
   const int kAltId = 172;
   const int kSZFormatId = 173;
   const int kSetGainFId = 190;
   const int kSetGainValId = 191;
   const int kSetGainUnitId = 192;
   const int kSetGainSignId = 193;
   const int kSetGainId = 194;
   const int kSetGainCurrentId = 195;
   const int kSetGainCurrentUnitId = 196;
   const int kAddId = 200;
   const int kAddGainId = 200;
   const int kAddZpkId = 201;
   const int kAddPolyId = 202;
   const int kAddResGainId = 203;
   const int kAddNotchId = 204;
   const int kAddEllipId = 205;
   const int kAddButterId = 206;
   const int kAddCheby1Id = 207;
   const int kAddCheby2Id = 208;
   const int kAddCombId = 209;
   const int kAddSosId = 210;
   const int kAddZRootId = 211;
   const int kAddDirectId = 212;


   const int kB_BODEPLOT = 15;
   const int kB_STEPPLOT = 16;
   const int kB_ROOTPLOT = 17;
   const int kB_WRITEFILTER = 18;
   const int kB_WRITELOAD = 19;
   const int kB_OK = 20;
   const int kB_CANCEL = 21;


   const int kM_FILE_SAVELOAD = 130;
   const int kM_FILE_READONLY = 131;
   const int kM_FILE_OK = 132;
   const int kM_FILE_CANCEL = 133;
   const int kM_PLOT_BODE = 410;
   const int kM_PLOT_STEP = 411;
   const int kM_PLOT_IMPULSE = 412;
   const int kM_PLOT_RAMP = 413;
   const int kM_PLOT_SROOTS = 414;
   const int kM_PLOT_ZROOTS = 415;
   const int kM_PLOT_CLOSEDLOOP = 416;
   const int kM_PLOT_NEGCONV = 417;



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Layout grids							        //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   Int_t kModSelGrid[] = {
   440, 250,             // total width & height
   15, 25, 70, 24,       // Path label
   15, 185, 70, 24,      // File label
   15, 215, 70, 24,      // Module label
   85, 25, 300, 24,      // Path 
   395, 25, 24, 24,      // Directory up button
   85, 55, 335, 121,     // Directory list
   85, 185, 335, 24,     // File 
   85, 215, 335, 24,     // Module 
   0};

   Int_t kSecSelGrid[] = {
   270, 250,             // total width & height
   15, 25, 70, 24,       // select label
   85, 25, 100, 24,      // multiple section selection
   200, 25, 24, 24,      // select all button
   230, 25, 24, 24,      // select none button
   15, 65, 20, 24,       // 1st section
   35, 65, 90, 24,       // 1st section name
   15, 95, 20, 24,       // 2nd section
   35, 95, 90, 24,       // 2nd section name
   15, 125, 20, 24,      // 3rd section
   35, 125, 90, 24,      // 3rd section name
   15, 155, 20, 24,      // 4th section
   35, 155, 90, 24,      // 4th section name
   15, 185, 20, 24,      // 5th section
   35, 185, 90, 24,      // 5th section name
   145, 65, 20, 24,      // 6th section
   165, 65, 90, 24,      // 6th section name
   145, 95, 20, 24,      // 7th section
   165, 95, 90, 24,      // 7th section name
   145, 125, 20, 24,     // 8th section
   165, 125, 90, 24,     // 8th section name
   145, 155, 20, 24,     // 9th section
   165, 155, 90, 24,     // 9th section name
   145, 185, 20, 24,     // 10th section
   165, 185, 90, 24,     // 10th section name
   0};

   Int_t kSwitchGrid[] = {
   220, 250,             // total width & height
   15,   25,  70, 24,    // input label
   15,   55,  70, 24,    // output label
   15,   95,  70, 24,    // ramp time label
   15,  125,  70, 24,    // tolerance label
   15,  155,  70, 24,    // timeout label
   180,  95,  25, 24,    // unit label
   180, 155,  25, 24,    // unit label
   85,   25, 120, 24,    // input switching
   85,   55, 120, 24,    // output switching
   85,   95,  90, 24,    // ramp time
   85,  125,  90, 24,    // tolerance
   85,  155,  90, 24,    // timeout   
   0};

   Int_t kDesignGrid[] = {
   946, 310,             // total width & height
   15,   25,  70, 24,    // fsample label
   180,  25,  30, 24,    // unit label
   15,   55,  70, 24,    // command label
   15,  145,  70, 24,    // alternate label
   85,  205,  70, 24,    // s-plane label
   540, 205,  70, 24,    // z-plane label
   15,  235,  70, 24,    // add label
   15,  275,  70, 24,    // set gain label
   180, 275,  30, 24,    // unit label
   220, 275,  50, 24,    // value label
   625, 275,  60, 24,    // curren value label
   85,   25,  90, 24,    // sample rate
   85,   55, 845, 84,    // command
   85,  145, 845, 54,    // alternate
   150, 205,  70, 24,    // s-plane format: s
   235, 205,  70, 24,    // s-plane format: f
   305, 205,  85, 24,    // s-plane format: n
   405, 205,  80, 24,    // s-plane format: p
   605, 205,  60, 24,    // z-plane format: SOS
   675, 205,  60, 24,    // z-plane format: roots
   745, 205,  60, 24,    // z-plane format: direct
   820, 205,  79, 24,    // z-plane format: online
   85,  275,  90, 24,    // set gain f
   265, 275,  90, 24,    // set gain value
   365, 275,  50, 24,    // set gain unit
   425, 275,  25, 24,    // set gain sign
   470, 275,  77, 24,    // set gain
   680, 275, 165, 24,    // set gain current value
   853, 275,  76, 24,    // set gain current value unit
   85,  235,  65, 24,    // gain
   150, 235,  65, 24,    // ZPK
   215, 235,  65, 24,    // poly
   280, 235,  65, 24,    // resgain
   345, 235,  65, 24,    // notch
   410, 235,  65, 24,    // elliptic
   475, 235,  65, 24,    // butter
   540, 235,  65, 24,    // cheby1
   605, 235,  65, 24,    // cheby2
   670, 235,  65, 24,    // comb
   735, 235,  65, 24,    // sos
   800, 235,  65, 24,    // zroots
   865, 235,  65, 24,    // direct
   930, 235,  60, 24,    // test
   0};

   Int_t kPlotGrid[] = {
   946, 60,              // total width & height
   15,  25, 40, 24,      // start label
   175, 25, 40, 24,      // stop label
   335, 25, 100,24,      // points label
   565, 25, 40, 24,      // type label
   780, 25, 50, 24,      // duration label
   55,  25, 90, 24,      // start frequency
   215, 25, 90, 24,      // stop frequency
   445, 25, 90, 24,      // number of points
   605, 25, 110,24,      // plot type
   840, 25, 90, 24,      // time duration
   0};



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Utility function			                                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   inline static double power (double x, double y) 
   {
      return exp (log (x) * y);
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGGridLayout						        //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   class TLGGridLayout : public TGLayoutManager {
   public:
      TLGGridLayout (TGCompositeFrame* p, const Int_t* coord) 
      : fMain (p), fCoord (coord) {
         fList = fMain->GetList(); }
      virtual void Layout ();
      virtual TGDimension GetDefaultSize() const {
         return TGDimension (fCoord[0], fCoord[1]); }
   
   protected:
      TGCompositeFrame*	fMain;
      TList*		fList;
      const Int_t*	fCoord;
   };

//______________________________________________________________________________
   void TLGGridLayout::Layout ()
   {
      const Int_t* i = fCoord + 2;
      TGFrameElement* ptr;
      TIter next (fList);
      while ((ptr = (TGFrameElement*) next())) {
         ptr->fFrame->MoveResize (i[0], i[1], i[2], i[3]);
         i += 4;
      }
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGFilterWizWindow                                                   //
//                                                                      //
// Filter wizard frame                                                  //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   TLGFilterWizWindow::TLGFilterWizWindow (const TGWindow* parent, 
                     std::string& fdesign, bool simple, 
                     const char* path, const char* file, Int_t id)
   : TGVerticalFrame (parent, 10, 10), TGWidget (id), fSimple (simple),
   fReadOnly (kTRUE), fClosedLoop (kFALSE), fClosedLoopNeg (kFALSE),
   fFilter (&fdesign), fCurModule (0), fCurSectionActive (kFALSE), 
   fCurSwitchActive (kFALSE), fCurDesignActive (kFALSE), fDirty (kFALSE)
   {
      for (int i = 0; i < 6; ++i) fF[i] = 0;
      for (int i = 0; i < 5; ++i) fG[i] = 0;
      for (int i = 0; i < 30; ++i) fLabel[i] = 0;
      fPath = 0;
      fCdup = 0;
      fPcdup = 0;
      fFv = 0;
      fFc = 0;
      fFile = 0;
      fModule = 0;
      fSectMul = 0;
      fPSectSel[0] = 0;
      fPSectSel[1] = 0;
      fSectSel[0] = 0;
      fSectSel[1] = 0;
      for (int i = 0; i < kMaxFilterSections; ++i) {
         fSect[i] = 0;
         fSectName[i] = 0;
      }
      fSwitchInp = 0;
      fSwitchOut = 0;
      for (int i = 0; i < 3; ++i) fSwitchParam[i] = 0;
   
      // layouts
      fL[0] = new TGLayoutHints (kLHintsLeft | kLHintsExpandX |
                           kLHintsTop, 0, 0, 0, 0);
      fL[1] = new TGLayoutHints (kLHintsLeft | kLHintsTop, 4, 4, 4, 4);
      fL[2] = new TGLayoutHints (kLHintsLeft | kLHintsExpandX | 
                           kLHintsTop, 4, 4, 4, 4);
      fL[3] = new TGLayoutHints (kLHintsRight | kLHintsCenterY, 6, 4, 2, 2);
      fL[4] = new TGLayoutHints (kLHintsLeft | kLHintsExpandX |
                           kLHintsTop, 0, 0, 2, unsigned(-6));
      fL[5] = new TGLayoutHints (kLHintsLeft | kLHintsCenterY, 2, 2, 2, 2);
      fL[6] = new TGLayoutHints (kLHintsLeft | kLHintsCenterY, 4, 
				 unsigned(-2), 2, 2);
      fL[7] = new TGLayoutHints (kLHintsLeft | kLHintsExpandX |
                           kLHintsTop, 0, 0, 2, 2);
      // groups
      if (!fSimple) {
         fF[0] = new TGHorizontalFrame (this, 10, 10);
         AddFrame (fF[0], fL[0]);
         fG[0] = new TGGroupFrame (fF[0], "Module Selection");
         fF[0]->AddFrame (fG[0], fL[1]);
         fG[1] = new TGGroupFrame (fF[0], "Sections");
         fF[0]->AddFrame (fG[1], fL[1]);
         fG[2] = new TGGroupFrame (fF[0], "Switching");
         fF[0]->AddFrame (fG[2], fL[1]);
      }
      fG[3] = new TGGroupFrame (this, "Design");
      AddFrame (fG[3], fL[2]);
      fG[4] = new TGGroupFrame (this, "Plotting");
      AddFrame (fG[4], fL[2]);
   
      // module selection
      if (!fSimple) {
         fG[0]->SetLayoutManager (new TLGGridLayout (fG[0], kModSelGrid));
         for (int i = 0; i < 3; ++i) {
            const char* const labels[] = 
               {"Path:", "File:", "Module:"};
            fLabel[i] = new TGLabel (fG[0], labels[i]);
            fLabel[i]->SetTextJustify (kTextLeft | kTextCenterY);
            fG[0]->AddFrame (fLabel[i], 0);
         }
         fPath = new TGFSComboBox (fG[0], kPathId);
         fPath->Associate (this);
         fG[0]->AddFrame (fPath, 0);
         fPcdup = fClient->GetPicture("tb_uplevel.xpm");
         if (!fPcdup) {
            Error("TLGFilterWizWindow", "missing toolbar pixmap(s).n");
         }
         fCdup = new TGPictureButton (fG[0], fPcdup, kDirUpId);
         fCdup->SetToolTipText("Up One Level");
         fCdup->Associate(this);
         fG[0]->AddFrame(fCdup, 0);
         fFv = new TGListView (fG[0], 400, 161);
         fFc = new TGFileContainer(fFv->GetViewPort(),
                              10, 10, kHorizontalFrame, fgWhitePixel);
         fFc->Associate (this);
         fFv->GetViewPort()->SetBackgroundColor (fgWhitePixel);
         fFv->SetContainer (fFc);
         fFv->SetViewMode (kLVList);
         fFc->SetFilter ("YOUneverHAVEaFILElikeTHIS!!!");
         fFc->Sort (kSortByType);
         if (!path) {
            path = ".";
         }
         fFc->ChangeDirectory (path);
         fPath->Update(fFc->GetDirectory());
         fG[0]->AddFrame(fFv, 0);
         fFile = new TGComboBox (fG[0], kFileId);
         fFile->Associate (this);
         fG[0]->AddFrame (fFile, 0);
         fModule = new TGComboBox (fG[0], kModuleId);
         fModule->Associate (this);
         fG[0]->AddFrame (fModule, 0);
      }
   
      // sections selection
      if (!fSimple) {
         fG[1]->SetLayoutManager (new TLGGridLayout (fG[1], kSecSelGrid));
         fLabel[3] = new TGLabel (fG[1], "Select:");
         fLabel[3]->SetTextJustify (kTextLeft | kTextCenterY);
         fG[1]->AddFrame (fLabel[3], 0);
         fSectMul = new TGComboBox (fG[1], kSectMulId);
         fSectMul->Associate (this);
         fSectMul->AddEntry ("Single", 0);
         fSectMul->AddEntry ("Multiple", 1);
         fSectMul->Select (0);
         fG[1]->AddFrame (fSectMul, 0);
         fPSectSel[0] = fClient->GetPicture ("arrow_up.xpm");
         fPSectSel[1] = fClient->GetPicture ("arrow_down.xpm");
         if (!fPSectSel[0] || !fPSectSel[1]) {
            Error("TLGFilterWizWindow", "missing toolbar pixmap(s).n");
         }
         fSectSel[0] = new TGPictureButton (fG[1], fPSectSel[0], 
                              kSectSelAllId);
         fSectSel[0]->SetToolTipText("Select All");
         fSectSel[0]->Associate(this);
         fG[1]->AddFrame (fSectSel[0], 0);
         fSectSel[1] = new TGPictureButton (fG[1], fPSectSel[1], 
                              kSectSelNoneId);
         fSectSel[1]->SetToolTipText("Deselect All");
         fSectSel[1]->Associate(this);
         fG[1]->AddFrame (fSectSel[1], 0);
         for (int i = 0; i < kMaxFilterSections; ++i) {
            char buf[64];
            sprintf (buf, "Filter Section %i", i);
            fSect[i] = new TGCheckButton (fG[1], "", kSectId + i);
            fSect[i]->SetToolTipText (buf);
            fSect[i]->Associate(this);
            fG[1]->AddFrame (fSect[i], 0);
            fSectName[i] = new TLGTextEntry (fG[1], "", kSectNameId + i);
            fSectName[i]->Associate(this);
            fG[1]->AddFrame (fSectName[i], 0);
         }
         fSect[0]->SetState (kButtonDown);
         fCurSections.insert (0);
      }
   
      // Switching
      if (!fSimple) {
         fG[2]->SetLayoutManager (new TLGGridLayout (fG[2], kSwitchGrid));
         for (int i = 0; i < 7; ++i) {
            const char* const labels[] = 
               {"Input:", "Output:", "Ramp Time:", "Tolerance:", 
               "Timeout:", "sec", "sec"};
            fLabel[i+4] = new TGLabel (fG[2], labels[i]);
            fLabel[i+4]->SetTextJustify (kTextLeft | kTextCenterY);
            fG[2]->AddFrame (fLabel[i+4], 0);
         }
         fSwitchInp = new TLGComboBox (fG[2], kSwitchInpId);
         fSwitchInp->Associate (this);
         fG[2]->AddFrame (fSwitchInp, 0);
         fSwitchOut = new TLGComboBox (fG[2], kSwitchOutId);
         fSwitchOut->Associate (this);
         fG[2]->AddFrame (fSwitchOut, 0);
         for (int i = 0; i < 3; ++i) {
            fSwitchParam[i] = new TLGNumericControlBox (fG[2], 
                                 0.0, 6, kSwitchParamId + i, kNESReal,
                                 kNEANonNegative);
            fSwitchParam[i]->Associate (this);
            fSwitchParam[i]->SetState (kFALSE);
            fG[2]->AddFrame (fSwitchParam[i], 0);
         }
      }
   
      // Plotting
      fG[4]->SetLayoutManager (new TLGGridLayout (fG[4], kPlotGrid));
      for (int i = 0; i < 5; ++i) {
         const char* const labels[] = 
            {"fStart:", "fStop:", "Number of Points:", "Type:",
            "Duration:"};
         fLabel[i+11] = new TGLabel (fG[4], labels[i]);
         fLabel[i+11]->SetTextJustify (kTextLeft | kTextCenterY);
         fG[4]->AddFrame (fLabel[i+11], 0);
      }
      for (int i = 0; i < 3; ++i) {
         const double numinit[] = {1, 1000, 1001};
         fPlotParam[i] = new TLGNumericControlBox (fG[4], 
                              numinit[i], 6, kPlotParamId + i,
                              i == 2 ? kNESInteger : kNESReal,
                              kNEAPositive);
         fPlotParam[i]->Associate (this);
         fG[4]->AddFrame (fPlotParam[i], 0);
      }
      fPlotType = new TGComboBox (fG[4], kPlotTypeId);
      fPlotType->Associate (this);
      fPlotType->AddEntry ("Logarithmic", 0);
      fPlotType->AddEntry ("Linear", 1);
      fPlotType->Select (0);
      fG[4]->AddFrame (fPlotType, 0);
      fPlotParam[3] = new TLGNumericControlBox (fG[4], 1.0, 6, 
                           kPlotParamId + 3, kNESReal, kNEAPositive);
      fPlotParam[3]->Associate (this);
      fG[4]->AddFrame (fPlotParam[3], 0);
   
      // Design
      fG[3]->SetLayoutManager (new TLGGridLayout (fG[3], kDesignGrid));
      //--------  Labels
      for (int i = 0; i < 11; ++i) {
         const char* const labels[] = 
            {"fSample:", "Hz", "Command:", "Alternate:", "s-Plane:", 
            "z-Plane:", "Add:", "Gain:", "Hz", "Value:",
            "Current:" };
         fLabel[i+16] = new TGLabel (fG[3], labels[i]);
         fLabel[i+16]->SetTextJustify (kTextLeft | kTextCenterY);
         fG[3]->AddFrame (fLabel[i+16], 0);
      }

      //--------  Sample rate selection
      fSample = new TLGNumericControlBox (fG[3], 
                           2048, 6, kSampleId, kNESReal,
                           kNEAPositive);
      fSample->Associate (this);
      fG[3]->AddFrame (fSample, 0);

      //--------  Command text-box
      fCommand = new TLGTextEdit (fG[3], 10, 10, kCommandId);
      fCommand->Associate (this);
      fG[3]->AddFrame (fCommand, 0);

      //--------  Alternate text-box
      fAlt = new TGTextView (fG[3], 10, 10, kAltId);
      fAlt->Associate (this);
      fG[3]->AddFrame (fAlt, 0);

      //--------  Mode raio buttonos
      for (int i = 0; i < 8; ++i) {
         const char* const labels[] = 
            {"s (rad/s)", "f (Hz)", "n (Hz/norm)", "Polynomial",
            "SOS", "Roots", "Direct", "Online"};
         const char* const tooltips[] = 
            {"Format with poles in left half plane and units of rad/s", 
            "Format with poles in left half plane and units of Hz", 
            "Format with poles in left half plane, units of Hz and "
            "normalized gains",
            "Rational polynomial in s", 
            "Second Order Section format", 
            "Roots of the z-plane",
            "Direct form",
            "Format used by online system"}; 
         fSZFormat[i] = new TGRadioButton (fG[3], labels[i], 
                              kSZFormatId + i);
         fSZFormat[i]->SetToolTipText (tooltips[i]);
         fSZFormat[i]->Associate (this);
         fG[3]->AddFrame (fSZFormat[i], 0);
      }
      fSZFormat[2]->SetState (kButtonDown);
      fSetGainF = new TLGNumericControlBox (fG[3], 
                           0, 6, kSetGainFId, kNESReal,
                           kNEAPositive);
      fSetGainF->Associate (this);
      fG[3]->AddFrame (fSetGainF, 0);
      fSetGainVal = new TLGNumericControlBox (fG[3], 
                           1, 6, kSetGainValId, kNESReal);
      fSetGainVal->Associate (this);
      fG[3]->AddFrame (fSetGainVal, 0);
      fSetGainUnit = new TLGComboBox (fG[3], kSetGainUnitId);
      fSetGainUnit->Associate (this);
      fSetGainUnit->AddEntry ("mag", 0);
      fSetGainUnit->AddEntry ("dB", 1);
      fSetGainUnit->Select (0);
      fG[3]->AddFrame (fSetGainUnit, 0);
      fSetGainSign = new TGTextButton (fG[3], "=", kSetGainSignId);
      fSetGainSign->SetToolTipText 
         ("Sign of real part of transfer coefficient");
      fSetGainSign->Associate (this);
      fG[3]->AddFrame (fSetGainSign, 0);
      fSetGain = new TGTextButton (fG[3], "Set", kSetGainId);
      fSetGain->SetToolTipText 
         ("Set the gain at frequency f to the specified value");
      fSetGain->Associate(this);
      fG[3]->AddFrame (fSetGain, 0);
      fSetGainCurrent = new TGTextEntry (fG[3], "", kSetGainCurrentId);
      fSetGainCurrent->Associate (this);
      fSetGainCurrent->SetState (kFALSE);
      fG[3]->AddFrame (fSetGainCurrent, 0);
      fSetGainCurrentUnit = new TLGComboBox (fG[3], kSetGainCurrentUnitId);
      fSetGainCurrentUnit->Associate (this);
      fSetGainCurrentUnit->AddEntry ("dB/deg", 0);
      fSetGainCurrentUnit->AddEntry ("mag/deg", 1);
      fSetGainCurrentUnit->AddEntry ("Re/Im", 2);
      fSetGainCurrentUnit->AddEntry ("dB/rad", 3);
      fSetGainCurrentUnit->AddEntry ("mag/rad", 4);
      fSetGainCurrentUnit->Select (0);
      fG[3]->AddFrame (fSetGainCurrentUnit, 0);
      // buttons
      for (int i = 0; i < 13; ++i) {
         const char* const labels[] = 
            {"Gain", "ZPK", "RPoly", "ResGain", "Notch",
            "Ellip", "Butter", "Cheby1", "Cheby2", 
            "Comb", "SOS", "zRoots", "Direct"};
         const char* const tooltips[] = 
            {"Gain multiplier", "Zero-pole-gain specification", 
            "Rational polynomial", "Resonant gain filter", "Notch filter", 
            "Elliptic filter", "Butterworth filter", "Chebyshev Type 1 filter", 
            "Chebyshev Type 2 filter",
            "Comb filter (notch or resonant gain)",
            "List of second order sections", 
            "List of z-plane roots", "Direct form (avoid using)"}; 
         fAdd[i] = new TGTextButton (fG[3], labels[i], kAddId + i);
         fAdd[i]->SetToolTipText (tooltips[i]);
         fAdd[i]->Associate(this);
         fG[3]->AddFrame (fAdd[i], 0);
      }
   
      Setup (path, file);
   }

//______________________________________________________________________________
   TLGFilterWizWindow::~TLGFilterWizWindow ()
   {
      delete fSetGainCurrentUnit;
      delete fSetGainCurrent;
      delete fSetGain;
      delete fSetGainSign;
      delete fSetGainUnit;
      delete fSetGainVal;
      delete fSetGainF;
      for (int i = 0; i < 13; ++i) delete fAdd[i];
      for (int i = 0; i < 8; ++i) delete fSZFormat[i];
      delete fAlt;
      delete fCommand;
      delete fSample;
      delete fPlotType;
      for (int i = 0; i < 4; ++i) delete fPlotParam[i];
      for (int i = 0; i < 3; ++i) delete fSwitchParam[i];
      delete fSwitchOut;
      delete fSwitchInp;
      for (int i = 0; i < kMaxFilterSections; ++i) delete fSectName[i];
      for (int i = 0; i < kMaxFilterSections; ++i) delete fSect[i];
      for (int i = 0; i < 2; ++i) 
         if (fPSectSel[i]) fClient->FreePicture (fPSectSel[i]);
      for (int i = 0; i < 2; ++i) delete fSectSel[i];
      delete fSectMul;
      delete fModule;
      delete fFile;
      delete fFc;
      delete fFv;
      delete fCdup;
      if (fPcdup) fClient->FreePicture (fPcdup);
      delete fPath;
      for (int i = 0; i < 27; ++i) delete fLabel[i];
      for (int i = 0; i < 5; ++i) delete fG[i];
      for (int i = 0; i < 1; ++i) delete fF[i];
      for (int i = 0; i < 8; ++i) delete fL[i];
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::AddFiles (const char* dir,
                     const char* file)
   {
      typedef std::vector<std::string> stringlist;
   
      // loop over files
      void *dirp;
      if ((dirp = gSystem->OpenDirectory (dir)) == 0) {
         fFile->SetTopEntry 
            (new TGTextLBEntry (fFile, new TGString (""), 0), 
            new TGLayoutHints (kLHintsLeft | kLHintsExpandY | 
                              kLHintsExpandX));
         fFile->MapSubwindows();
         fFile->Layout();
         return kFALSE;
      }

      //-------------------------------- Scan file in directory. Compare
      //                                 first line to "magic string"
      stringlist nlist;
      int lMagic = strlen(kMagic);
      const char* name;
      while ((name = gSystem->GetDirEntry (dirp)) != 0) {
         if (strcmp (name, ".") && strcmp(name, "..")) {
            // found a file: open it and check magic token
            string fullname = string (dir) + "/" + name;
            ifstream inp (fullname.c_str());
            char line[128];
            inp.read (line, lMagic);
            if (inp && (strncmp (line, kMagic, lMagic) == 0)) {
               nlist.push_back (name);
            }
         }
      }
      gSystem->FreeDirectory(dirp);
      sort (nlist.begin(), nlist.end());


      int id = 0;
      int cur = -1;
      for (stringlist::iterator i = nlist.begin(); i != nlist.end(); ++i) {
         if (file && (*i == file)) cur = id;
         fFile->AddEntry (i->c_str(), id++);
      }
      if (cur < 0) cur = 0;
      // select file
      if (id > 0) {
         fFile->MapSubwindows();
         fFile->Select (cur);
      }
      else {
         fFile->SetTopEntry 
            (new TGTextLBEntry (fFile, new TGString (""), 0), 
            new TGLayoutHints (kLHintsLeft | kLHintsExpandY | 
                              kLHintsExpandX));
         fFile->MapSubwindows();
         fFile->Layout();
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::Setup (const char* dir, const char* file,
                     const char* module)
   {
      if (fSimple) {
         fFilterFile.add ("Default", 16384);
         fCurModule = fFilterFile.find ("Default");
         fCurSections.insert (0);
         if (fCurModule && fFilter) {
            (*fCurModule)[0].setDesign (fFilter->c_str());
         }
         return kTRUE;
      }
      if (dir && (fCurDir == TString (dir))) {
         return kTRUE;
      }
      fFile->RemoveEntries (0, 10000);
      if (!dir) {
         ReadFile (0);
         return kFALSE;
      }
      if (!AddFiles (dir, file)) {
         ReadFile (0);
         return kTRUE;
      }
      fCurDir = dir;
      TGTextLBEntry* e = (TGTextLBEntry*)fFile->GetSelectedEntry();
      if (!e) {
         ReadFile (0);
         return kTRUE;
      }
      return ReadFile (e->GetText()->GetString(), module);
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateFileSelection()
   {
      if (fSimple) {
         return kTRUE;
      }
      fFc->ChangeDirectory (fCurDir);
      fPath->Update (fFc->GetDirectory());
      fFile->RemoveEntries (0, 10000);
      if (!AddFiles (fCurDir, fCurFile)) {
         ReadFile (0);
         return kFALSE;
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::ReadOnly() const
   {
      return fReadOnly;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::SetReadOnly (Bool_t readonly)
   {
      if (readonly != fReadOnly) {
         fReadOnly = readonly;
         Update (kTRUE, kTRUE);
      }
      return kTRUE;
   }

//______________________________________________________________________________
   void TLGFilterWizWindow::SetDirty (Bool_t on)
   {
      fDirty = on;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::CheckDirty()
   {
      if (!IsDirty() || ReadOnly()) {
         return kTRUE;
      }
      Int_t ret;
      new TGMsgBox (gClient->GetRoot(), this, "Save", 
                   "Do you want to save changes?",
                   kMBIconQuestion, kMBYes | kMBNo | kMBCancel, &ret);
      if (ret == kMBCancel) {
         return kFALSE;
      }
      else if (ret == kMBYes) {
         return SaveFile();
      }
      else {
         return kTRUE;
      }
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::IsValid (Bool_t msgbox)
   {
      if (fCurFile == "") {
         return kTRUE;
      }
      string errmsg;
      if ((fCurFile == "") || fFilterFile.valid (errmsg)) {
         return kTRUE;
      }
      if (msgbox) {
         errmsg += "Invalid filter modules/sections:\n" + errmsg;
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      errmsg.c_str(), kMBIconExclamation, kMBOk, 0);
      }
      return kFALSE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::ReadFile (const char* file, 
                     const char* module)
   {
      cerr << "READ FILE " << (file ? file : "") << endl;
      if (file && (fCurFile == TString (file))) {
         return kTRUE;
      }
      string oldname = fCurModule ? fCurModule->getName() : "";
      fModule->RemoveEntries (0, 10000);
      string fullname = string (fCurDir) + "/" + (file ? file : "");
      if (!file || !fFilterFile.read (fullname.c_str())) {
         fCurModule = 0;
         fCurFile = "";
         SetDirty (kFALSE);
         return SelectModule (0);
      }
      fCurModule = 0;
      fCurFile = file;
      SetDirty (kFALSE);
      // Add modules to selection list
      int cur = -1;
      int id = 0;
      for (FilterModuleList::iterator i = fFilterFile.modules().begin();
          i != fFilterFile.modules().end(); ++i) {
         if (module && strcmp (module, i->getName()) == 0) cur = id;
         if ((cur < 0) && !oldname.empty() &&
            strcmp (oldname.c_str(), i->getName()) == 0) cur = id;
         fModule->AddEntry (i->getName(), id++);
      }
      fModule->MapSubwindows();
      fModule->Layout();
      if (cur < 0) cur = 0;
      // Set module
      if (id > 0) {
         fCurModule = 0;
         fModule->Select (cur);
         return SelectModule (((TGTextLBEntry*)fModule->GetSelectedEntry())->
                             GetText()->GetString());
      }
      return SelectModule (0);
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::SaveFile()
   {
      if (ReadOnly()) {   
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to save filters in readonly mode.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      if (!IsDirty()) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Nothing to save.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kTRUE;
      }
      if (!IsValid()) {
         return kFALSE;
      }
      string filename = (const char*)fCurFile;
      if (filename.empty()) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Cannot save empty file.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      Update (kFALSE);
      // save
      filename = string (fCurDir) + "/" + filename;
      if (!fFilterFile.write (filename.c_str())) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to write file.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      SetDirty (kFALSE);
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::SaveAs()
   {
      if (!IsValid()) {
         return kFALSE;
      }
      // file save as dialog
      TGFileInfo info;
      info.fFilename = 0;
      info.fIniDir = 0;
   #if ROOT_VERSION_CODE >= ROOT_VERSION(3,1,6)
      info.fFileTypes = const_cast<const char**>(gSaveAsTypes);
   #else
      info.fFileTypes = const_cast<char**>(gSaveAsTypes);
   #endif
      if (!TLGFileDialog (this, info, kFDSave, ".txt")) {
         return kFALSE;
      }
      string filename = info.fFilename ? info.fFilename : "";
   #if ROOT_VERSION_CODE < ROOT_VERSION(3,1,6)
      delete [] info.fFilename;
   #endif
      // save file
      Update (kFALSE);
      if (!fFilterFile.write (filename.c_str())) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to write file.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      // update file selection box
      fCurDir = gSystem->DirName (filename.c_str());
      fCurFile = gSystem->BaseName (filename.c_str());
      UpdateFileSelection();
      SetDirty (kFALSE);
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::LoadCoeff ()
   {
      if (ReadOnly()) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to load filter coefficients in readonly "
                      "mode.", kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      return kFALSE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::SetGain (double f, double g, 
                     sign_type sign)
   {
      if (!fCurModule || ReadOnly() || (fCurSections.size() != 1) ||
         !Update (kFALSE)) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to set filter gain.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      FilterSection& sect = (*fCurModule)[*fCurSections.begin()];
   
      // make filter and get gain
      FilterDesign ds (fCurModule->getFSample());
      fComplex coeff;
      if (!ds.filter (sect.getDesign()) || !ds.Xfer (coeff, f)) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to set filter gain.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      // adjust gain
      double gain_adj = 1;
      if (coeff.Mag() > 1E-12) gain_adj = fabs (g) / coeff.Mag();
      if (gain_adj < 1E-12) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to set filter gain.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      // adjust sign
      int pos = true;
      switch (sign) {
         case kSignSame:
            {
               break;
            }
         case kSignMinus:
            pos = false;
            // no break!
         case kSignPlus:
            {
               bool coeff_pos = (coeff.Real() != 0) ? 
                  (coeff.Real() > 0) : (coeff.Imag() >= 0);
               if (coeff_pos != pos) {
                  gain_adj *= -1;
               }
               break;
            }
      }
   
      // add gain
      if (fabs (gain_adj - 1.0) < 1E-12) {
         return kTRUE;
      }
      string cmd = sect.getDesign();
      char buf[256];
      sprintf (buf, "gain(%g)", gain_adj);
      cmd += buf;
      sect.setDesign (cmd.c_str());
      // update and return
      return UpdateDesign (kTRUE, kTRUE);
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::SelectModule (const char* module)
   {
      if (fSimple) {
         return kTRUE;
      }
      Update (kFALSE);
      if (!module) {
         fCurModule = 0;
         fModule->SetTopEntry 
            (new TGTextLBEntry (fModule, new TGString (""), 0), 
            new TGLayoutHints (kLHintsLeft | kLHintsExpandY | 
                              kLHintsExpandX));
         fModule->MapSubwindows();
         Update (kTRUE, kTRUE);
         return kTRUE;
      }
      if (fCurModule && 
         (strcmp (fCurModule->getName(), module) == 0)) {
         return kTRUE;
      }
      fCurModule = fFilterFile.find (module);
      if (!fCurModule) {
         fModule->SetTopEntry 
            (new TGTextLBEntry (fModule, new TGString (""), 0), 
            new TGLayoutHints (kLHintsLeft | kLHintsExpandY | 
                              kLHintsExpandX));
         fModule->MapSubwindows();
         Update();
         return kFALSE;
      }
      return Update();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::SelectSections()
   {
      if (fSimple) {
         return kTRUE;
      }
      UpdateSwitching (kFALSE);
      UpdateDesign(kFALSE);
      fCurSections.clear();
      for (int i = 0; i < kMaxFilterSections; ++i) {
         if (fSect[i]->GetState() == kButtonDown) {
            fCurSections.insert (i);
         }
      }
      return UpdateSwitching() && UpdateDesign();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateSections (Bool_t toGUI, Bool_t force)
   {
      if (fSimple) {
         return kTRUE;
      }
      // do we have a module selected?
      if (!fCurModule) {
         if (fCurSectionActive || force) {
            for (int i = 0; i < kMaxFilterSections; ++i) {
               fSectName[i]->SetText ("");
               fSectName[i]->SetState (kFALSE);
            }
            fCurSectionActive = kFALSE;
         }
         return kTRUE;
      }
      // make active if inactive
      if (toGUI && (!fCurSectionActive || force)) {
         for (int i = 0; i < kMaxFilterSections; ++i) {
            fSectName[i]->SetState (ReadOnly() ? kFALSE : kTRUE);
         }
         fCurSectionActive = kTRUE;
      }
      // transfer data
      if (toGUI) {
         for (int i = 0; i < kMaxFilterSections; ++i) {
            fSectName[i]->SetText ((*fCurModule)[i].getName());
         }
      }
      else {
         for (int i = 0; i < kMaxFilterSections; ++i) {
            (*fCurModule)[i].setName(fSectName[i]->GetText ());
         }
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateSwitchParams()
   {
      if (fSimple) {
         return kTRUE;
      }
      int sel = fSwitchOut->GetSelected();
      switch (sel) {
         // Immediately
         case 1:
            {
               fSwitchParam[0]->SetState (kFALSE);
               fSwitchParam[1]->SetState (kFALSE);
               fSwitchParam[2]->SetState (kFALSE);
               break;
            }
         // Ramp
         case 2:
            {
               fSwitchParam[0]->SetState 
                  (ReadOnly() ? kFALSE : kTRUE);
               fSwitchParam[1]->SetState (kFALSE);
               fSwitchParam[2]->SetState (kFALSE);
               break;
            }
         // Input crossing
         case 3:
            {
               fSwitchParam[0]->SetState (kFALSE);
               fSwitchParam[1]->SetState 
                  (ReadOnly() ? kFALSE : kTRUE);
               fSwitchParam[2]->SetState 
                  (ReadOnly() ? kFALSE : kTRUE);
               break;
            }
         // Zero crossing
         case 4:
            {
               fSwitchParam[0]->SetState (kFALSE);
               fSwitchParam[1]->SetState 
                  (ReadOnly() ? kFALSE : kTRUE);
               fSwitchParam[2]->SetState 
                  (ReadOnly() ? kFALSE : kTRUE);
               break;
            }
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateSwitching (Bool_t toGUI, 
                     Bool_t force)
   {
      if (fSimple) {
         return kTRUE;
      }
      // do we have multiple sections selected?
      if ((fCurSections.size() != 1) || !fCurModule) {
         if (fCurSwitchActive || force) {
            fSwitchInp->RemoveEntries (0, 10000);
            fSwitchOut->RemoveEntries (0, 10000);
            fSwitchInp->SetTopEntry 
               (new TGTextLBEntry (fSwitchInp, new TGString (""), 0), 
               new TGLayoutHints (kLHintsLeft | kLHintsExpandY | 
                                 kLHintsExpandX));
            fSwitchInp->MapSubwindows();
            fSwitchOut->SetTopEntry 
               (new TGTextLBEntry (fSwitchOut, new TGString (""), 0), 
               new TGLayoutHints (kLHintsLeft | kLHintsExpandY | 
                                 kLHintsExpandX));
            fSwitchOut->MapSubwindows();
            fSwitchInp->SetState (kFALSE);
            fSwitchOut->SetState (kFALSE);
            for (int i = 0; i < 3; ++i) {
               fSwitchParam[0]->SetState (kFALSE);
            }
            fCurSwitchActive = kFALSE;
         }
         return kTRUE;
      }
      // make active if inactive
      if (toGUI && !fCurSwitchActive) {
         fSwitchInp->AddEntry ("Always On", kAlwaysOn);
         fSwitchInp->AddEntry ("Zero History", kZeroHistory);
         fSwitchOut->AddEntry ("Immediately", 1);
         fSwitchOut->AddEntry ("Ramp", 2);
         fSwitchOut->AddEntry ("Input Crossing", 3);
         fSwitchOut->AddEntry ("Zero Crossing", 4);
         fCurSwitchActive = kTRUE;
      }
      // transfer data
      FilterSection& sect = (*fCurModule)[*fCurSections.begin()];
      if (toGUI) {
         fSwitchInp->Select ((int)sect.getInputSwitch());
         fSwitchOut->Select ((int)sect.getOutputSwitch());
         fSwitchParam[0]->SetNumber (sect.getRamp());
         fSwitchParam[1]->SetNumber (sect.getTolerance());
         fSwitchParam[2]->SetNumber (sect.getTimeout());
         fSwitchInp->SetState (ReadOnly() ? kFALSE : kTRUE);
         fSwitchOut->SetState (ReadOnly() ? kFALSE : kTRUE);
         UpdateSwitchParams();
      }
      else {
         sect.setInputSwitch ((input_switching)
                             fSwitchInp->GetSelected());
         sect.setOutputSwitch ((output_switching)
                              fSwitchOut->GetSelected());
         sect.setRamp (fSwitchParam[0]->GetNumber());
         sect.setTolerance (fSwitchParam[1]->GetNumber());
         sect.setTimeout (fSwitchParam[2]->GetNumber());
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateDesign (Bool_t toGUI,
                     Bool_t force)
   {
      // do we have a module selected?
      if (!fCurModule) {
         fSample->SetState (kFALSE);
         if (fCurDesignActive || force) {
            fCommand->LoadBuffer (" ");
            fCommand->SetState (kFALSE);
            fAlt->LoadBuffer (" ");
            fSetGainVal->SetState (kFALSE);
            fSetGainUnit->SetState (kFALSE);
            fSetGainSign->SetState (kButtonDisabled);
            fSetGain->SetState (kButtonDisabled);
            fCurDesignActive = kFALSE;
         }
         return kTRUE;
      }
      // check active
      fSample->SetState (ReadOnly() ? kFALSE : kTRUE);
      if ((fCurSections.size() == 1) && 
         (!fCurDesignActive || force)) {
         fCommand->SetState (ReadOnly() ? kFALSE : kTRUE);
         fSetGainVal->SetState (ReadOnly() ? kFALSE : kTRUE);
         fSetGainUnit->SetState (ReadOnly() ? kFALSE : kTRUE);
         fSetGainSign->SetState (ReadOnly() ? kButtonDisabled : kButtonUp);
         fSetGain->SetState (ReadOnly() ? kButtonDisabled : kButtonUp);
         fCurDesignActive = kTRUE;
      }
      if ((fCurSections.size() != 1) && 
         (fCurDesignActive || force)) {
         fCommand->LoadBuffer (" ");
         fCommand->SetState (kFALSE);
         fAlt->LoadBuffer (" ");
         fSetGainVal->SetState (kFALSE);
         fSetGainUnit->SetState (kFALSE);
         fSetGainSign->SetState (kButtonDisabled);
         fSetGain->SetState (kButtonDisabled);
         fCurDesignActive = kFALSE;
      }
      // transfer data
      if (toGUI) {
         fSample->SetNumber (fCurModule->getFSample());
         string design;
         for (section_sel::iterator i = fCurSections.begin(); 
             i != fCurSections.end(); ++i) {
            const char* d = (*fCurModule)[*i].getDesign();
            if (d && *d && !design.empty()) design += "*";
            if (d && *d) design += d;
         }
         design = FilterSection::splitCmd (design.c_str());
         if (design.empty()) design = " ";
         fCommand->LoadBuffer (design.c_str());
         UpdateDesignZP();
      }
      else if (fCurSections.size() == 1) {
         fCurModule->setFSample (fSample->GetNumber());
         if (fCurSections.size() == 1) {
            string cmd = fCommand->GetString();
            (*fCurModule)[*fCurSections.begin()].setDesign (cmd.c_str());
         }
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateDesignZP()
   {
      double fsample = fSample->GetNumber();
      string cmd = fCommand->GetString('#');
      string plane = "s";
      string format = "s";
      string alt;
      bool splane = true;
      if (fSZFormat[0]->GetState() == kButtonDown) {
         plane = "s";
      }
      else if (fSZFormat[1]->GetState() == kButtonDown) {
         plane = "f";
      }
      else if (fSZFormat[2]->GetState() == kButtonDown) {
         plane = "n";
      }
      else if (fSZFormat[3]->GetState() == kButtonDown) {
         plane = "p";
      }
      else if (fSZFormat[4]->GetState() == kButtonDown) {
         format = "s";
         splane = false;
      }
      else if (fSZFormat[5]->GetState() == kButtonDown) {
         format = "r";
         splane = false;
      }
      else if (fSZFormat[6]->GetState() == kButtonDown) {
         format = "d";
         splane = false;
      }
      else if (fSZFormat[7]->GetState() == kButtonDown) {
         format = "o";
         splane = false;
      }
      else {
         fSetGainCurrent->SetText ("");
         return kFALSE;
      }
      // make a filter
      FilterDesign ds (fsample);
      if (!ds.filter (cmd.c_str())) {
         fAlt->LoadBuffer ("Error");
         fSetGainCurrent->SetText ("");
         return kFALSE;
      }
      // set alternate form
      if (splane) {
         if (!iir2zpk (ds(), alt, plane.c_str())) {
            fAlt->LoadBuffer ("Error");
            fSetGainCurrent->SetText ("");
            return kFALSE;
         }
         else {
            alt = FilterSection::splitCmd (alt.c_str());
            if (alt.empty()) alt = " ";
            if ((alt == "zpk([],[],1)") || (alt == "zpk([],[],1,\"f\")") ||
               (alt == "zpk([],[],1,\"n\")")) {
               alt = " ";
            }
            fAlt->LoadBuffer (alt.c_str());
         }
      }
      else {
         if (!iir2z (ds(), alt, format.c_str())) {
            fAlt->LoadBuffer ("Error");
            fSetGainCurrent->SetText ("");
            return kFALSE;
         }
         else {
            alt = FilterSection::splitCmd (alt.c_str());
            if (alt.empty()) alt = " ";
            if ((alt == "zroots([],[],1)") || (alt == "sos(1,[])") ||
               (alt == "sos(1,[],\"o\")")) {
               alt = " ";
            }
            fAlt->LoadBuffer (alt.c_str());
         }
      }
      // set current gain value
      char buf[256];
      fComplex coeff;
      double f = fSetGainF->GetNumber();
      int unit = fSetGainCurrentUnit->GetSelected();
      if (ds.Xfer (coeff, f)) {
         double n1 = 0;
         double n2 = 0;
         switch (unit) {
            // dB/deg
            case 0:
               {
                  n1 = 20. * log10 (coeff.Mag());
                  n2 = 180. / pi * coeff.Arg();
                  break;
               }
            // mag/deg
            case 1:
               {
                  n1 = coeff.Mag();
                  n2 = 180. / pi * coeff.Arg();
                  break;
               }
            // Real/Imag
            case 2:
               {
                  n1 = coeff.Real();
                  n2 = coeff.Imag();
                  break;
               }
            // dB/rad
            case 3:
               {
                  n1 = 20. * log10 (coeff.Mag());
                  n2 = coeff.Arg();
                  break;
               }
            // mag/rad
            case 4:
               {
                  n1 = coeff.Mag();
                  n2 = coeff.Arg();
                  break;
               }
         }
         sprintf (buf, "%g  %g", n1, n2);
      }
      else {
         strcpy (buf, "");
      }
      fSetGainCurrent->SetText (buf);
   
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::Update (Bool_t toGUI, Bool_t force)
   {
      return UpdateSections (toGUI, force) && 
         UpdateSwitching (toGUI, force) && 
         UpdateDesign (toGUI, force);
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::UpdateFilter()
   {
      if (!fSimple || !fFilter) {
         return kFALSE;
      }
      Update (kFALSE);
      *fFilter = (*fCurModule)[0].getDesign();
      return kTRUE;
   }

//______________________________________________________________________________
   PlotDescriptor* TLGFilterWizWindow::PlotBode ()
   {
      PlotDescriptor* const retNull(0);

      if (!fCurModule || !Update (kFALSE)) {
         return retNull;
      }
      // get sample & cmd
      double fsample = fSample->GetNumber();
      string cmd = fCommand->GetString('#');
      // get sweep parameters
      double start = fPlotParam[0]->GetNumber();
      double stop = fPlotParam[1]->GetNumber();
      int points = fPlotParam[2]->GetIntNumber();
      string sweeptype = "log";
      if (fPlotType->GetSelected() == 1) {
         sweeptype = "linear";
      }
      if (points < 1) points = 1;
      if (stop < start) swap (start, stop);
      if (stop <= 0) {
         return retNull;
      }
      if (start < 0) start = 0;
      if ((start == 0) && (sweeptype == "log")) {
         start = 1E-3;
      }
      // add filter
      FilterDesign ds (fsample);
      if (!ds.filter (cmd.c_str())) {
         return retNull;
      }
      if (fClosedLoop) {
         if (!ds.closeloop (fClosedLoopNeg ? -1 : +1)) {
            return retNull;
         }
      }
      // Compute Xfer
      DataDescriptor* data = new DataCopy (0, 0, points, true);
      if (!ds.Xfer (data->GetX(), (fComplex*)data->GetY(), 
                   start, stop, points, sweeptype.c_str())) {
         delete data;
         return retNull;
      }
      // Make plot descriptor
      string name = fCurModule->getName();
      if (fCurSections.size() == 1) {
         char buf[16];
         sprintf (buf, "%i", *fCurSections.begin());
         name += buf;
      }
      else if ((int)fCurSections.size() != kMaxFilterSections) {
         for (section_sel::iterator i = fCurSections.begin(); 
             i != fCurSections.end(); ++i) {
            char buf[16];
            sprintf (buf, "_%i", *i);
            name += buf;
         }   
      }
      string achn = name + "_in";
      string bchn = name + "_out";
      ParameterDescriptor param;
      Time now (Now().getS(), 0);
      param.SetStartTime (now.getS(), now.getN());
      param.SetThird (fCurFile);
      calibration::Descriptor caldesc (now, kPTTransferFunction, 
                           achn.c_str(), bchn.c_str());
      PlotDescriptor* pd = 
         new PlotDescriptor (data, kPTTransferFunction, achn.c_str(), 
                           bchn.c_str(), &param, &caldesc);
   
      return pd;
   }

//______________________________________________________________________________
   PlotDescriptor* TLGFilterWizWindow::PlotResponse (const char* resp)
   {
      PlotDescriptor* const retNull(0);

      if (!fCurModule || !Update (kFALSE)) {
         return retNull;
      }
      // get sample & cmd
      double fsample = fSample->GetNumber();
      string cmd = fCommand->GetString('#');
      // get time duration
      double duration = fPlotParam[3]->GetNumber();
      if (duration < 1. / fsample) duration = 1.0;
   
      // add filter
      FilterDesign ds (fsample);
      if (!ds.filter (cmd.c_str())) {
         return retNull;
      }
      if (fClosedLoop) {
         if (!ds.closeloop (fClosedLoopNeg ? -1 : +1)) {
            return retNull;
         }
      }
   
      // Compute response
      TSeries out (Time (0,0), Interval (0), 1);
      if (!ds.response (out, resp, Interval (duration))) {
         return retNull;
      }
      int points = out.getNSample();
      DataDescriptor* data = new DataCopy (0., out.getTStep(), 0, points);
      out.getData (points, data->GetY());
      // Make plot descriptor
      string name = fCurModule->getName();
      if (fCurSections.size() == 1) {
         char buf[16];
         sprintf (buf, "%i", *fCurSections.begin());
         name += buf;
      }
      else if ((int)fCurSections.size() != kMaxFilterSections) {
         for (section_sel::iterator i = fCurSections.begin(); 
             i != fCurSections.end(); ++i) {
            char buf[16];
            sprintf (buf, "_%i", *i);
            name += buf;
         }   
      }
      name += string ("_") + resp;
      ParameterDescriptor param;
      Time now (Now().getS(), 0);
      param.SetStartTime (now.getS(), now.getN());
      param.SetThird (fCurFile);
      calibration::Descriptor caldesc (now, kPTTimeSeries, 
                           name.c_str());
      PlotDescriptor* pd = 
         new PlotDescriptor (data, kPTTimeSeries, name.c_str(), 
                           0, &param, &caldesc);
   
      return pd;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::PlotRoots (PlotDescriptor*& poles,
                     PlotDescriptor*& zeros, bool splane)
   {
      if (!fCurModule || !Update (kFALSE)) {
         return 0;
      }
      // get sample & cmd
      double fsample = fSample->GetNumber();
      string cmd = fCommand->GetString('#');
      // add filter
      FilterDesign ds (fsample);
      if (!ds.filter (cmd.c_str())) {
         return kFALSE;
      }
      if (fClosedLoop) {
         if (!ds.closeloop (fClosedLoopNeg ? -1 : +1)) {
            return kFALSE;
         }
      }
      // get poles and zeros
      int order = iirorder (ds.get());
      dComplex* p = new dComplex [order + 1];
      dComplex* z = new dComplex [order + 1];
      int npoles;
      int nzeros;
      double g;
      if ((splane && !iir2zpk (ds.get(), nzeros, z, npoles, p, g, "f")) ||
         (!splane && !iir2z (ds.get(), nzeros, z, npoles, p, g))) {
         delete [] p;
         delete [] z;
         return kFALSE;
      }
      // make data descriptors
      DataDescriptor* p_data = new DataCopy (0, 0, npoles);
      for (int i = 0; i < npoles; ++i) {
         p_data->GetX()[i] = p[i].Real();
         p_data->GetY()[i] = p[i].Imag();
      }
      DataDescriptor* z_data = new DataCopy (0, 0, nzeros);
      for (int i = 0; i < nzeros; ++i) {
         z_data->GetX()[i] = z[i].Real();
         z_data->GetY()[i] = z[i].Imag();
      }
      delete [] p;
      delete [] z;
      // create XY plot descriptor
      string name = fCurModule->getName();
      if (fCurSections.size() == 1) {
         char buf[16];
         sprintf (buf, "%i", *fCurSections.begin());
         name += buf;
      }
      else if ((int)fCurSections.size() != kMaxFilterSections) {
         for (section_sel::iterator i = fCurSections.begin(); 
             i != fCurSections.end(); ++i) {
            char buf[16];
            sprintf (buf, "_%i", *i);
            name += buf;
         }   
      }
      name += string ("_");
      ParameterDescriptor param;
      Time now (Now().getS(), 0);
      param.SetStartTime (now.getS(), now.getN());
      param.SetThird (fCurFile);
      string a = name + "poles_real";
      string b = name + "poles_imag";
      calibration::Descriptor p_caldesc (now, kPTXY, 
                           a.c_str(), b.c_str());
      if (splane) {
         p_caldesc.SetDomain (kCalUnitX, kCalDomainFrequency);
         p_caldesc.SetDomain (kCalUnitY, kCalDomainFrequency);
      }
      poles = new PlotDescriptor (p_data, kPTXY, a.c_str(), 
                           b.c_str(), &param, &p_caldesc);
      a = name + "zeros_real";
      b = name + "zeros_imag";
      calibration::Descriptor z_caldesc (now, kPTXY, 
                           a.c_str(), b.c_str());
      if (splane) {
         z_caldesc.SetDomain (kCalUnitX, kCalDomainFrequency);
         z_caldesc.SetDomain (kCalUnitY, kCalDomainFrequency);
      }
      zeros = new PlotDescriptor (z_data, kPTXY, a.c_str(), 
                           b.c_str(), &param, &z_caldesc);
      return kTRUE;
   }

//______________________________________________________________________________
   PlotDescriptor* TLGFilterWizWindow::UnitCircle (int segments)
   {
      // make data descriptors
      int n = segments + 1;
      DataDescriptor* data = new DataCopy (0, 0, n);
      for (int i = 0; i < segments + 1; ++i) {
         dComplex z;
         z.setMArg (1., 2. * pi * (double)i / (double)segments);
         data->GetX()[i] = z.Real();
         data->GetY()[i] = z.Imag();
      }
      // create XY plot descriptor
      string name = "unit_circle_";
      ParameterDescriptor param;
      Time now(0);
      param.SetStartTime (0, 0);
      param.SetThird ("Unit Circle");
      string a = name + "real";
      string b = name + "imag";
      calibration::Descriptor p_caldesc (now, kPTXY, 
                           a.c_str(), b.c_str());
      PlotDescriptor* pd = 
         new PlotDescriptor (data, kPTXY, a.c_str(), 
                           b.c_str(), &param, &p_caldesc);
      return pd;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::AddFilter (Long_t parm1)
   {
      // check for readonly
      if (ReadOnly()) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Cannot add filters in readonly mode.", 
                      kMBIconExclamation, kMBOk, 0);
         return kFALSE;
      }
      // check that a filter module is selected
      if (!fCurModule) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Need to have a module selected for "
                      "adding a filter.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      // Check only one section is selcted
      if (fCurSections.size() != 1) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Exactly one filter section has to be "
                      "selected for adding a filter.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      // check that we have a valid design
      if (!UpdateDesign (kFALSE) ||
         (!(*fCurModule)[*fCurSections.begin()].update())) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Must start with a valid filter.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      // get marked string
      TString cmd;
      if (fCommand->IsMarked()) {
         fCommand->Copy();
         TGText* clip = fCommand->GetClipboard();
         for (int i = 0; i < clip->RowCount(); ++i) {
            char* p = clip->GetLine 
               (TGLongPosition (0, i), clip->ColCount());
            if (p) {
               if (cmd != "") cmd += "\n";
               cmd += p;
            }
            delete [] p;
         }
         // check marked string
         FilterDesign ds (fCurModule->getFSample());
         if (!ds.filter (cmd)) {
            new TGMsgBox (gClient->GetRoot(), this, "Error", 
                         "Selection must be a valid filter.", 
                         kMBIconExclamation, kMBOk, 0);         
            return kFALSE;
         }
      }
   
      // Bring up corresponding dialog box
      TLGNotchDialog::subtype_t sub1 = TLGNotchDialog::kNotch;
      TLGEllipDialog::subtype_t sub2 = TLGEllipDialog::kEllip;
      switch (parm1) {
         // Gain factor
         case kAddGainId:
            {
               new TLGGainDialog (gClient->GetRoot(), this, cmd);
               break;
            }
         // Poles and zeros
         case kAddZpkId:
         case kAddZRootId:
            {
               double fsample = fSample->GetNumber();
               new TLGZpkDialog (gClient->GetRoot(), this, fsample,
                                cmd, parm1 == kAddZpkId);
               break;
            }
         // Polynomial forms
         case kAddPolyId:
         case kAddDirectId:
            {
               double fsample = fSample->GetNumber();
               new TLGPolyDialog (gClient->GetRoot(), this, 
                                 cmd, fsample, parm1 == kAddPolyId);
               break;
            }
         // Notch, resonant gain and comb
         case kAddCombId:
            sub1 = TLGNotchDialog::kComb;
         // no break;
         case kAddResGainId:
            if (sub1 == TLGNotchDialog::kNotch)
               sub1 = TLGNotchDialog::kResGain;
         // no break;
         case kAddNotchId: 
            {
               new TLGNotchDialog (gClient->GetRoot(), this, 
                                  cmd, sub1);
               break;
            }
         // Elliptic, butterworth, chebyshev
         case kAddButterId:
            sub2 = TLGEllipDialog::kButter;
         // no break;
         case kAddCheby1Id:
            if (sub2 == TLGEllipDialog::kEllip)
               sub2 = TLGEllipDialog::kCheby1;
         // no break;
         case kAddCheby2Id:
            if (sub2 == TLGEllipDialog::kEllip)
               sub2 = TLGEllipDialog::kCheby2;
         // no break;
         case kAddEllipId:
            {
               new TLGEllipDialog (gClient->GetRoot(), this, 
                                  cmd, sub2);
               break;
            }
         // SOS
         case kAddSosId:
            {
               double fsample = fSample->GetNumber();
               new TLGSosDialog (gClient->GetRoot(), this, 
                                fsample, cmd);
               break;
            }
         // ???
         default:
            {
               new TGMsgBox (gClient->GetRoot(), this, "Error", 
                            "Unknown filter.", 
                            kMBIconExclamation, kMBOk, 0);         
               break;
            }
      }
      // cancel?
      if (cmd == "") {
         return kTRUE;
      }
      SetDirty();
      // cut out marked text and replace
      if (fCommand->IsMarked()) {
         TGText text (cmd);
         fCommand->Delete();
         fCommand->InsertText (&text, kTRUE);
         if (!UpdateDesign (kFALSE) || 
            !(*fCurModule)[*fCurSections.begin()].update()) {
            new TGMsgBox (gClient->GetRoot(), this, "Error", 
                         "Unable to add filter:\n" + cmd, 
                         kMBIconExclamation, kMBOk, 0); 
            return kFALSE;        
         }
         UpdateDesignZP();
      }
      // otherwise add at the end
      else {
         if (!(*fCurModule)[*fCurSections.begin()].add (cmd)) {
            new TGMsgBox (gClient->GetRoot(), this, "Error", 
                         "Unable to add filter:\n" + cmd, 
                         kMBIconExclamation, kMBOk, 0); 
            return kFALSE;        
         }
         UpdateDesign();
      }
   
      return kTRUE;
   }


//______________________________________________________________________________
   Bool_t TLGFilterWizWindow::ProcessMessage (Long_t msg, Long_t parm1, 
                     Long_t parm2)
   {
      // Buttons
      if ((GET_MSG (msg) == kC_COMMAND) &&
         (GET_SUBMSG (msg) == kCM_BUTTON)) {
         switch (parm1) {
            // Directory up
            case kDirUpId:
               {
                  if (!fSimple && CheckDirty()) {
                     fFc->ChangeDirectory ("..");
                     fPath->Update (fFc->GetDirectory());
                     Setup (fFc->GetDirectory());
                  }
                  break;     
               }
            // Select all sections
            case kSectSelAllId:
               {
                  if (!fSimple) {
                     fSectMul->Select (1);
                     for (int i = 0; i < kMaxFilterSections; ++i) {
                        fSect[i]->SetState (kButtonDown);
                     }
                     SelectSections();
                  }
                  break;
               }
            // Deselect all sections
            case kSectSelNoneId:
               {
                  if (!fSimple) {
                     fSectMul->Select (0);
                     for (int i = 0; i < kMaxFilterSections; ++i) {
                        fSect[i]->SetState (kButtonUp);
                     }
                     SelectSections();
                  }
                  break;
               }
            // Set gain sign
            case kSetGainSignId:
               {
                  if (fSetGainSign->GetString() == "=") {
                     fSetGainSign->SetText ("+");
                  }
                  else if (fSetGainSign->GetString() == "+") {
                     fSetGainSign->SetText ("-");
                  }
                  else {
                     fSetGainSign->SetText ("=");
                  }
                  break;
               }
            // Set gain
            case kSetGainId:
               {
                  double f = fSetGainF->GetNumber();
                  double g = fSetGainVal->GetNumber();
                  sign_type sign = kSignSame;
                  if (fSetGainSign->GetString() == "+") sign = kSignPlus;
                  if (fSetGainSign->GetString() == "-") sign = kSignMinus;
                  if (fSetGainUnit->GetSelected() == 1) {
                     g = power (10, g / 20.);
                  }
                  if (SetGain (f, fabs (g), sign)) {
                     SetDirty();
                  }
                  break;
               }
         }
         // Add filter buttons
         if ((parm1 >= kAddId) && (parm1 < kAddId + 20)) {
            AddFilter (parm1);
         }
      }
      // Checkbox
      else if ((GET_MSG (msg) == kC_COMMAND) &&
              (GET_SUBMSG (msg) == kCM_CHECKBUTTON)) {
         // Sections
         if (!fSimple && (parm1 >= kSectId) && 
            (parm1 < kSectId + kMaxFilterSections)) {
            if (fSectMul->GetSelected() == 0) {
               for (int i = 0; i < kMaxFilterSections; ++i) {
                  fSect[i]->SetState (i == parm1 - kSectId ?
                                     kButtonDown : kButtonUp);
               }
            }
            SelectSections();
         }
      }
      // Radio buttons
      else if ((GET_MSG (msg) == kC_COMMAND) &&
              (GET_SUBMSG (msg) == kCM_RADIOBUTTON)) {
         // s/z-Plane format
         if ((parm1 >= kSZFormatId) && (parm1 < kSZFormatId + 8)) {
            for (int i = 0; i < 8; ++i) {
               fSZFormat[i]->SetState 
                  (i == parm1 - kSZFormatId ? kButtonDown: kButtonUp);
            }
            UpdateDesignZP();
         }
      }
      // Comboboxes
      else if ((GET_MSG (msg) == kC_COMMAND) &&
              (GET_SUBMSG (msg) == kCM_COMBOBOX)) {
         switch (parm1) {
            // Select path
            case kPathId:
               {
                  if (!fSimple) {
                     TGTreeLBEntry* e = 
                        (TGTreeLBEntry*) fPath->GetSelectedEntry();
                     if (e && CheckDirty()) {
                        fFc->ChangeDirectory (e->GetPath()->GetString());
                        fPath->Update (fFc->GetDirectory());
                        Setup (fFc->GetDirectory());
                     }
                     else {
                        UpdateFileSelection();
                     }
                  }
                  break;
               }
            // Select file
            case kFileId:
               {
                  if (!fSimple) {
                     TGTextLBEntry* e = 
                        (TGTextLBEntry*)fFile->GetSelectedEntry();
                     if (e && CheckDirty()) {
                        ReadFile (e->GetText()->GetString());
                     }
                     else {
                        UpdateFileSelection();
                     }
                  }
                  break;
               }
            // Select module
            case kModuleId:
               {
                  if (!fSimple) {
                     TGTextLBEntry* e = 
                        (TGTextLBEntry*)fModule->GetSelectedEntry();
                     if (e) {
                        SelectModule (e->GetText()->GetString());
                     }
                  }
                  break;
               }
            // Multiple section select
            case kSectMulId:
               {
                  // if single, make sure only one is selected
                  if (!fSimple && fSectMul->GetSelected() == 0) {
                     bool firstup = false;
                     bool deselect = false;
                     for (int i = 0; i < kMaxFilterSections; ++i) {
                        if (firstup &&
                           (fSect[i]->GetState() == kButtonDown)) {
                           deselect = true;
                           fSect[i]->SetState (kButtonUp);
                        }
                        if (fSect[i]->GetState() == kButtonDown) {
                           firstup = true;
                        }
                     }
                     if (deselect) SelectSections();
                  }
                  break;
               }
            // Input switching
            case kSwitchInpId:
               {
                  SetDirty();
                  break;
               }
            // Output switching
            case kSwitchOutId:
               {
                  UpdateSwitchParams();
                  SetDirty();
                  break;
               }
            // Set gain current unit
            case kSetGainCurrentUnitId:
               {
                  UpdateDesignZP();
                  break;
               }
         }
      }
      // File container double click
      else if ((GET_MSG (msg) == kC_CONTAINER) &&
              (GET_SUBMSG (msg) == kCT_ITEMDBLCLICK)) {
         // left mouse selects dir
         if (!fSimple && (parm1 == kButton1)) {
            if ((fFc->NumSelected() == 1) && CheckDirty()) {
               void* p = 0;
               TGFileItem* f = (TGFileItem *) fFc->GetNextSelected (&p);
               fFc->ChangeDirectory (f->GetItemName()->GetString());
               fPath->Update (fFc->GetDirectory());
               Setup (fFc->GetDirectory());
            }
            else {
               UpdateFileSelection();
            }
         }
      }
      // numeric field updated
      else if ((GET_MSG (msg) == kC_TEXTENTRY) &&
              (GET_SUBMSG (msg) == kTE_TEXTUPDATED)) {
         switch (parm1) {
            // sample
            case kSampleId:
               {
                  UpdateDesignZP();
                  SetDirty();
                  break;
               }
            // command
            case kCommandId:
               {
                  UpdateDesignZP();
                  SetDirty();
                  break;
               }
            // set gain frequency
            case kSetGainFId:
               {
                  UpdateDesignZP();
                  break;
               }
         }
      }
      // text fields changed
      else if ((GET_MSG (msg) == kC_TEXTENTRY) &&
              (GET_SUBMSG (msg) == kTE_TEXTCHANGED)) {
         SetDirty();
      }
      return kTRUE;
   }



//////////////////////////////////////////////////////////////////////////
//                                                                      //
// TLGFilterWizard                                                      //
//                                                                      //
// Filter wizard main window                                            //
//                                                                      //
//////////////////////////////////////////////////////////////////////////
   TLGFilterWizard::TLGFilterWizard (const TGWindow *p, bool& modal,
                     const char* title, std::string& fdesign,
                     const char* path, const char* file)
   : TLGMainWindow (p, title), fModal (modal), fReturn (&modal),
   fFilter (&fdesign), fIniPath (path), fIniFile (file)
   {
   }

//______________________________________________________________________________
   TLGFilterWizard::~TLGFilterWizard ()
   {
      delete fFilterDesign;
      if (fMPad) delete fMPad;
      fMPad = 0;
      delete fMainTab;
      delete fFilterLayout;
   }

//______________________________________________________________________________
   void TLGFilterWizard::CloseWindow()
   {
      if (fModal) {
         DeleteWindow();
      }
      else {
         TLGMainWindow::CloseWindow();
      }
   }

//______________________________________________________________________________
   int TLGFilterWizard::AddButtons (TGCompositeFrame* p, TGButton* btns[], 
                     int max, TGLayoutHints* btnLayout,
                     GContext_t btnGC, FontStruct_t btnFont) 
   {
      int num = 0;
      TGButton* tmp_btn[max];

      if (fModal) {
         const char* const buttontext[] = 
            {"Bode Plot", "Step Response", "s-Plane Roots", "Ok", 
            "Cancel"};
         const char* const buttonhelp[] = 
            {"Plot the transfer function of the filter",
            "Plot the step response of the filter",
            "Plot the location of poles and zero in the s-plane",
            "Quit",
            "Exit without change"};
         const int buttonid[] = 
            {kB_BODEPLOT, kB_STEPPLOT, kB_ROOTPLOT, kB_OK, kB_CANCEL};

	 num = 5;
         for (int i = 0; i < num; i++) {
            tmp_btn[i] = new TGTextButton (p, buttontext[i], buttonid[i], 
                                 btnGC, btnFont);
            tmp_btn[i]->Associate (this);
            tmp_btn[i]->SetToolTipText (buttonhelp[i]);
            p->AddFrame (tmp_btn[i], btnLayout);
         }
      }
      else {
         const char* const buttontext[] = 
            {"Bode Plot", "Step Response", "s-Plane Roots", "Save", 
            "Load Coefficients"};
         const char* const buttonhelp[] = 
            {"Plot the transfer function of the filter",
            "Plot the step response of the filter",
            "Plot the location of poles and zero in the s-plane",
            "Save filters settings to file", 
            "Save filters to file and load coefficients into online system"};
         const int buttonid[] = 
            {kB_BODEPLOT, kB_STEPPLOT, kB_ROOTPLOT, kB_WRITEFILTER,
            kB_WRITELOAD};

	 num = 5;
         for (int i = 0; i < num; i++) {
            tmp_btn[i] = new TGTextButton (p, buttontext[i], buttonid[i], 
                                 btnGC, btnFont);
            tmp_btn[i]->Associate (this);
            tmp_btn[i]->SetToolTipText (buttonhelp[i]);
            if ((i == 3) || (i == 4)) {
               tmp_btn[i]->SetState (kButtonDisabled);
            }
            p->AddFrame (tmp_btn[i], btnLayout);
         }
      }

      //--------------------------------  Add the exit button.
      int nstd = AddStdButtons (ligogui::kButtonExit, 
				p, btns, max, btnLayout, btnGC, btnFont);

      for (int i = 0; i < num; i++) {
	btns[i] = tmp_btn[i];
      }

      //--------------------------------  The button layout was set here, but..
      //delete fButtonLayout;
      //fButtonLayout =
      //      new TGLayoutHints (kLHintsExpandX | kLHintsTop, 15, 15, 10, 6);
      return num + nstd;
   }

//______________________________________________________________________________
   TLGMultiPad* TLGFilterWizard::AddMainWindow (TGCompositeFrame* p,
                     TGLayoutHints* mainLayout,
                     PlotSet& plotset, int padnum)
   {
      fMainTab = new TGTab (p, 10, 10);
      fMainTab->Associate (this);
      p->AddFrame (fMainTab, mainLayout);
      // add filter design tab
      fFilterLayout = new TGLayoutHints (kLHintsCenterX | kLHintsCenterY, 
                           4, 4, 4, 4);
      fFilterTab = fMainTab->AddTab (" Design ");
      fFilterDesign = new TLGFilterWizWindow (fFilterTab, 
                           *fFilter, fModal, fIniPath, fIniFile, 
                           kMainFrameId + 1);
      fFilterDesign->Associate (this);
      fFilterTab->AddFrame (fFilterDesign, fFilterLayout);
      // add grpahics tab
      fGraphicsTab = fMainTab->AddTab (" Graphics ");
      if (padnum < 1) padnum = 1;
      TLGMultiPad* pad = new TLGMultiPad (fGraphicsTab, "Plot", plotset, 
                           kMainFrameId, padnum);
      pad->Associate (this);
      if (padnum > 0) pad->GetPad(0)->HidePanel (kFALSE);
      if (padnum > 1) pad->GetPad(1)->HidePanel (kFALSE);
      fGraphicsTab->AddFrame (pad, mainLayout);
      return pad;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::ReadOnly() const
   {
      return fFilterDesign->ReadOnly();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::SetReadOnly (Bool_t readonly)
   {
      fFilterDesign->SetReadOnly (readonly);
      bool succ = fFilterDesign->SetReadOnly (readonly);
      if (succ && !fModal) {
         if (ReadOnly()) {
            fMenuFile->CheckEntry (kM_FILE_READONLY);
            fButtons[3]->SetState (kButtonDisabled);
            fButtons[4]->SetState (kButtonDisabled);
         }
         else {
            fMenuFile->UnCheckEntry (kM_FILE_READONLY);
            fButtons[3]->SetState (kButtonUp);
            fButtons[4]->SetState (kButtonDisabled); // for now
            //fButtons[4]->SetState (kButtonUp);
         }
      }
      return succ;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::FileSave ()
   {
      if (ReadOnly()) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to save filters in readonly mode.", 
                      kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      return fFilterDesign->SaveFile();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::FileSaveAs ()
   {
      return fFilterDesign->SaveAs();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::LoadCoeff ()
   {
      if (ReadOnly()) {
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to load filter coefficients in readonly "
                      "mode.", kMBIconExclamation, kMBOk, 0);         
         return kFALSE;
      }
      return fFilterDesign->LoadCoeff();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::BodePlot ()
   {
      PlotDescriptor* pd = fFilterDesign->PlotBode();
      if (!pd) {
         Int_t ret;
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      "Unable to generate Bode plot.", kMBIconExclamation, 
                      kMBOk, &ret);
         return kFALSE;
      }
      if (fCalTable) fCalTable->AddUnits (pd->Cal());
      PlotDescriptor* cur = pd->Clone (pd->GetGraphType(),
                           "current_in", "current_out", fCalTable);
      fPlot->Add (cur);
      ShowDefaultPlot (kFALSE, 0, 1);
      fPlot->Add (pd);
      fPlot->Update();
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::ResponsePlot (const char* resp)
   {
      PlotDescriptor* pd = fFilterDesign->PlotResponse (resp);
      if (!pd) {
         Int_t ret;
         string msg = string ("Unable to generate ") + string (resp) +
            " response.";
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      msg.c_str(), kMBIconExclamation, 
                      kMBOk, &ret);
         return kFALSE;
      }
      if (fCalTable) fCalTable->AddUnits (pd->Cal());
      string name = string ("current_") + resp;
      PlotDescriptor* cur = pd->Clone (pd->GetGraphType(),
                           name.c_str(), 0, fCalTable);
      fPlot->Add (cur);
      ShowDefaultPlot (kFALSE, 0, 3);
      fPlot->Add (pd);
      fPlot->Update();
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::RootPlot (bool splane)
   {
      PlotDescriptor* pole_pd = 0;
      PlotDescriptor* zero_pd = 0;
      if (!fFilterDesign->PlotRoots (pole_pd, zero_pd, splane)) {
         Int_t ret;
         string msg = string ("Unable to generate plot of roots.");
         new TGMsgBox (gClient->GetRoot(), this, "Error", 
                      msg.c_str(), kMBIconExclamation, 
                      kMBOk, &ret);
         return kFALSE;
      }
      if (fCalTable) fCalTable->AddUnits (pole_pd->Cal());
      if (fCalTable) fCalTable->AddUnits (zero_pd->Cal());
      string a = string ("current_poles_real");
      string b = string ("current_poles_imag");
      PlotDescriptor* pole_cur = pole_pd->Clone (pole_pd->GetGraphType(),
                           a.c_str(), b.c_str(), fCalTable);
      a = string ("current_zeros_real");
      b = string ("current_zeros_imag");
      PlotDescriptor* zero_cur = zero_pd->Clone (zero_pd->GetGraphType(),
                           a.c_str(), b.c_str(), fCalTable);
      fPlot->Add (pole_cur);
      fPlot->Add (zero_cur);
      // setup plot
      OptionAll_t* opt = fMPad->GetPlotOptions (0);
      if ((opt != 0) && 
         ((opt->fTraces.fGraphType.Length() == 0) ||
         opt->fConfig.fAutoConf && !opt->fConfig.fRespectUser)) {
         if (opt->fConfig.fAutoConf) {
            SetDefaultGraphicsOptions (*opt);
         }
         // Defaults
         for (Int_t tr = 0; tr < kMaxTraces; tr++) {
            opt->fTraces.fPlotStyle[tr] = kPlotStyleMarker;
            // poles
            if (tr == 0) {
               opt->fTraces.fGraphType = pole_cur->GetGraphType();
               opt->fTraces.fAChannel[tr] = pole_cur->GetAChannel();
               if (pole_cur->GetBChannel() != 0) {
                  opt->fTraces.fBChannel[tr] = pole_cur->GetBChannel();
               }
               opt->fTraces.fActive[tr] = kTRUE;
            }
            // zeros
            else if (tr == 1) {
               opt->fTraces.fAChannel[tr] = zero_cur->GetAChannel();
               if (zero_cur->GetBChannel() != 0) {
                  opt->fTraces.fBChannel[tr] = zero_cur->GetBChannel();
               }
               opt->fTraces.fActive[tr] = kTRUE;
            }
            // add unit circle
            else if ((tr == kMaxTraces - 1) && !splane) {
               PlotDescriptor* unitc = TLGFilterWizWindow::UnitCircle();
               fPlot->Add (unitc);
               opt->fTraces.fGraphType = unitc->GetGraphType();
               opt->fTraces.fAChannel[tr] = unitc->GetAChannel();
               if (unitc->GetBChannel() != 0) {
                  opt->fTraces.fBChannel[tr] = unitc->GetBChannel();
               }
               opt->fTraces.fPlotStyle[tr] = kPlotStyleLine;
               opt->fTraces.fActive[tr] = kTRUE;
            }
            else {
               opt->fTraces.fActive[tr] = kFALSE;
            }
         }
         opt->fUnits.fYValues = kUnitMagnitude;
         opt->fRange.fAxisScale[0] = kAxisScaleLinear;
         opt->fRange.fAxisScale[1] = kAxisScaleLinear;
         opt->fRange.fRange[0] = kRangeAutomatic;
         opt->fRange.fRange[1] = kRangeAutomatic;
         opt->fRange.fBinLogSpacing = kFALSE;
         opt->fAxisX.fAxisTitle = splane ? "Real" : "Real";
         opt->fAxisY.fAxisTitle = splane ? "Imaginary" : "Imaginary";
         opt->fStyle.fTitle = splane ? "s-Plane" : "z-Plane";
         opt->fUnits.fXUnit = "Default";
         opt->fUnits.fYUnit = "Default";
         opt->fUnits.fXMag = 0;
         opt->fUnits.fYMag = 0;
         opt->fUnits.fXSlope = 1.;
         opt->fUnits.fXOffset = 0.;
         opt->fUnits.fYSlope = 1.;
         opt->fUnits.fYOffset = 0.;
      }
   
      fPlot->Add (pole_pd);
      fPlot->Add (zero_pd);
      fPlot->Update();
      return kFALSE;
   }

//______________________________________________________________________________
   void TLGFilterWizard::AddMenuFile()
   {
      if (!fMenuFile) {
         fMenuFile = new TGPopupMenu (gClient->GetRoot());
         fMenuFile->Associate (TLGMainMenu::fParent);
      }
      fMenuFile->AddEntry ("&New", kM_FILE_NEW);
      if (!fModal) {
         fMenuFile->AddEntry ("Read &Only", kM_FILE_READONLY);
         fMenuFile->AddSeparator();
         fMenuFile->AddEntry ("&Save", kM_FILE_SAVE);
         fMenuFile->AddEntry ("Save &As", kM_FILE_SAVEAS);
         fMenuFile->AddEntry ("Save and &Load", kM_FILE_SAVELOAD);
      }
      fMenuFile->AddSeparator();
      fMenuFile->AddEntry ("&Import...", kM_FILE_IMPORT);
      fMenuFile->AddEntry ("&Export...", kM_FILE_EXPORT);
      fMenuFile->AddSeparator();
      fMenuFile->AddEntry ("&Print...", kM_FILE_PRINT);
      fMenuFile->AddEntry ("P&rint Setup...", kM_FILE_PRINTSETUP);
      fMenuFilePrintGraph = new TGPopupMenu (gClient->GetRoot());
      fMenuFilePrintGraph->AddEntry ("A", kM_FILE_PRINT_GRAPHA);
      fMenuFilePrintGraph->AddEntry ("B", kM_FILE_PRINT_GRAPHB);
      fMenuFile->AddPopup ("Print &Graph", fMenuFilePrintGraph);
      fMenuFile->AddSeparator();
      if (!fModal) {
         fMenuFile->AddEntry ("E&xit", kM_FILE_EXIT);
      }
      else {
         fMenuFile->AddEntry ("Ok", kM_FILE_OK);
         fMenuFile->AddEntry ("Cancel", kM_FILE_CANCEL);
      }
      fMenuBar->AddPopup ("&File", fMenuFile, fMenuBarItemLayout);
      fMenuFile->CheckEntry (kM_FILE_READONLY);
   }

//______________________________________________________________________________
   void TLGFilterWizard::AddMenuPlot()
   {
      if (!fMenuPlot) {
         fMenuPlot = new TGPopupMenu (gClient->GetRoot());
         fMenuPlot->Associate (TLGMainMenu::fParent);
      }
      fMenuPlot->AddEntry ("&Bode Plot", kM_PLOT_BODE);
      fMenuPlot->AddEntry ("S&tep Response", kM_PLOT_STEP);
      fMenuPlot->AddEntry ("&Impulse Response", kM_PLOT_IMPULSE);
      fMenuPlot->AddEntry ("R&amp Response", kM_PLOT_RAMP);
      fMenuPlot->AddSeparator();
      fMenuPlot->AddEntry ("&s-Plane Roots", kM_PLOT_SROOTS);
      fMenuPlot->AddEntry ("&z-Plane Roots", kM_PLOT_ZROOTS);
      fMenuPlot->AddSeparator();
      fMenuPlot->AddEntry ("&Closed Loop", kM_PLOT_CLOSEDLOOP);
      fMenuPlot->AddEntry ("&1/(1-G(f))", kM_PLOT_NEGCONV);
      fMenuPlot->AddSeparator();
      TLGMainMenu::AddMenuPlot();
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::ProcessButton (Long_t parm1, Long_t parm2) 
   {
      switch (parm1) {
         case kB_BODEPLOT:
            {
               return BodePlot();
            }
         case kB_STEPPLOT:
            {
               return ResponsePlot ("step");
            }
         case kB_ROOTPLOT:
            {
               return RootPlot (true);
            }
         case kB_WRITEFILTER:
            {
               return FileSave();
            }
         case kB_WRITELOAD:
            {
               return FileSave() && LoadCoeff();
            }
         case kB_CANCEL:
            {
               if (fModal) {
                  if (fReturn) *fReturn = false;
                  DeleteWindow();
                  return kFALSE;
               }
               break;
            }
         case kB_OK:
            {
               if (fModal) {
                  bool succ = fFilterDesign->UpdateFilter();
                  if (fReturn) *fReturn = succ;
                  DeleteWindow();
                  return kTRUE;
               }
               break;
            }
         default:
            {
               return TLGMainWindow::ProcessButton (parm1, parm2);
            }
      }
      return kTRUE;
   }

//______________________________________________________________________________
   Bool_t TLGFilterWizard::ProcessMenu (Long_t parm1, Long_t parm2) 
   {
      switch (parm1) {
         case kM_FILE_READONLY:
            {
               return SetReadOnly (!ReadOnly());
            }
         case kM_FILE_SAVELOAD:
            {
               return FileSave() && LoadCoeff();
            }
         case kM_FILE_OK:
            {
               return ProcessButton (kB_OK, 0);
            }
         case kM_FILE_CANCEL:
            {
               return ProcessButton (kB_CANCEL, 0);
            }
         case kM_PLOT_BODE:
            {
               return BodePlot();
            }
         case kM_PLOT_STEP:
            {
               return ResponsePlot ("step");
            }
         case kM_PLOT_IMPULSE:
            {
               return ResponsePlot ("impulse");
            }
         case kM_PLOT_RAMP:
            {
               return ResponsePlot ("ramp");
            }
         case kM_PLOT_SROOTS:
            {
               return RootPlot (true);
            }
         case kM_PLOT_ZROOTS:
            {
               return RootPlot (false);
            }
         case kM_PLOT_CLOSEDLOOP:
            {
               fFilterDesign->SetClosedLoop (
                                    !fFilterDesign->ClosedLoop());
               if (fFilterDesign->ClosedLoop()) {
                  fMenuPlot->CheckEntry (kM_PLOT_CLOSEDLOOP);
               }
               else {
                  fMenuPlot->UnCheckEntry (kM_PLOT_CLOSEDLOOP);
               }
               return kTRUE;
            }
         case kM_PLOT_NEGCONV:
            {
               fFilterDesign->SetClosedLoopNeg (
                                    !fFilterDesign->ClosedLoopNeg());
               if (fFilterDesign->ClosedLoopNeg()) {
                  fMenuPlot->CheckEntry (kM_PLOT_NEGCONV);
               }
               else {
                  fMenuPlot->UnCheckEntry (kM_PLOT_NEGCONV);
               }
               return kTRUE;
            }
         case kM_HELP_ABOUT:
            {
               new TGMsgBox (gClient->GetRoot(), TLGMainMenu::fParent, 
                            "About", aboutmsg, 0, kMBOk);
               break;
            }
         default:
            {
               return TLGMainWindow::ProcessMenu (parm1, parm2);
            }
      }
      return kTRUE;
   }

}


//////////////////////////////////////////////////////////////////////////
//                                                                      //
// Extern                                                               //
//                                                                      //
// These functions are needed by the FilterDesign class			//
//                                                                      //
//////////////////////////////////////////////////////////////////////////
extern "C" {
   bool wizard___dynamic (const std::string name, std::string& filter)
   {
      if (!gClient) {
         return 0;
      }
      FilterDesign fdesign;
      string n = "Foton";
      if (!name.empty()) {
         n += ": ";
         n += name;
      }
   
      // create main window
      bool modal_ret = true;
      filterwiz::TLGFilterWizard* foton = new 
         filterwiz::TLGFilterWizard (gClient->GetRoot(), modal_ret, 
                           n.c_str(), filter, 0,  0);
      foton->Setup();
      foton->SetReadOnly (false);
      gClient->WaitFor (foton);
   
      return modal_ret;
   }

}
