#ifndef PTMON_HH
#define PTMON_HH

#include "DatEnv.hh"
#include "MonServer.hh"
#include "OperStateCondList.hh"
#include "TrigClient.hh"
#include "TrigPar.hh"
#include "PTChan.hh"
#include <list>
#include <map>
#include <string>
#include <vector>

//----------------------------------------- Include C headers & structures

#include "ligostd.h"
#include "detect.h"
#include "ligocont.h"
#include "ligoconfig.h"

//  PTmon is a glitch monitor based on the Peak-Trough time series algorithm 
//  developed by J.N. Murdock and C.R. Hutt for the US Geological Survey, and  
//  summarized in their Open File Report 83-785.  Further documentation is    
//  included in their Open File Report 87-158.  Both reports can be found at  
//  website of the Albuqurque Seismological Laboratory: //alswww.cr.usgs.gov/ 
//  under "ASL Publications".  The code has been adapted for use by LIGO by   
//  @LIGO_author N. Zotov
//  @version 1.0;  prepared January, 2003.

class PTmon : public DatEnv, MonServer, TrigClient {
public:
 
PTmon(int argc, const char *argv[]);           // Constructor
~PTmon();                                      // Destructor

    void ProcessData(void);
    void Attention(void);
    void Reset(void);
    bool ReadConfig(void);
    bool chanExists(const std::string& key) const;
 
    struct detect_info detect;
    struct s_onset onset;

    float repThresh;



private:

    typedef  std::map<std::string, PTChan> ChanList;
    typedef ChanList::iterator chan_iter;
    ChanList mChannels;

    typedef ChanList::const_iterator const_chan_iter;



    TrigPar mDict;
    double  mStride;
    double dt;
   
    const PTChan* mChan;
   
    std::string mCFile;
    OperStateCondList mOSC;
};


#endif    // !defined(PTMON_HH)
